import minio.error
import pytest
import urllib3

from .. import storage_clients


@pytest.mark.parametrize(
    "disable_cert_checks, debug",
    [(False, False), (False, True), (True, False), (True, True)],
)
def test_minio_disable_cert_checks(db, settings, disable_cert_checks, debug):
    settings.MINIO_DISABLE_CERT_CHECKS = disable_cert_checks
    settings.DEBUG = debug

    client = storage_clients.Minio()

    # Some hackery to avoid the test taking too long
    client.client._http.connection_pool_kw["retries"].total = 0

    if debug and disable_cert_checks:
        assert client.client._http.connection_pool_kw["cert_reqs"] == "CERT_NONE"

        # This should succeed, as we're not verifying the server certificate
        client.client._http.urlopen("get", "https://self-signed.badssl.com/")
    else:
        assert client.client._http.connection_pool_kw["cert_reqs"] == "CERT_REQUIRED"

        with pytest.raises(urllib3.exceptions.RequestError):
            # This should fail, as we are in fact verifying certificates
            client.client._http.urlopen("get", "https://self-signed.badssl.com/")


def _put_side_effect(*_, **__):
    # first call, raise an exception, second time, succeed
    yield minio.error.S3Error(
        "NoSuchBucket",
        "Bucket does not exist",
        resource="adsf",
        request_id="fake",
        host_id="fake",
        response="fake",
    )
    yield "upload_url_successful"


@pytest.fixture
def patched_minio(mocker):
    make_bucket = mocker.patch("minio.api.Minio.make_bucket")
    put_object = mocker.patch(
        "minio.api.Minio.presigned_put_object", side_effect=_put_side_effect()
    )
    return make_bucket, put_object


def test_minio_auto_create_bucket_enabled(db, settings, patched_minio):
    settings.MINIO_STORAGE_AUTO_CREATE_MEDIA_BUCKET = True
    client = storage_clients.Minio()

    make_bucket, _ = patched_minio

    assert client.upload_url("asdf") == "upload_url_successful"
    assert make_bucket.call_count == 1


def test_minio_auto_create_bucket_disabled(db, settings, patched_minio):
    settings.MINIO_STORAGE_AUTO_CREATE_MEDIA_BUCKET = False
    client = storage_clients.Minio()
    make_bucket, _ = patched_minio

    with pytest.raises(minio.error.S3Error):
        client.upload_url("asdf")

    assert make_bucket.call_count == 0


@pytest.mark.parametrize("exc_code", ["NoSuchBucket", "NoSuchKey"])
def test_minio_handle_exceptions(exc_code, caplog, mocker):
    # The "happy path" is tested in various places already
    mocker.patch(
        "minio.api.Minio.stat_object",
        side_effect=minio.error.S3Error(
            code=exc_code,
            message="",
            resource="test_object",
            request_id="",
            host_id="",
            response=None,
        ),
    )
    client = storage_clients.Minio()
    stat = client.stat_object("test_object")
    assert stat is None
    assert caplog.messages == [f"Minio error, cannot stat object: {exc_code}"]



from django.db import transaction
from rest_framework import exceptions
from rest_framework.serializers import (
    CharField,
    DateField,
    FloatField,
    IntegerField,
    ListField,
    PrimaryKeyRelatedField,
)

from ..caluma_core import serializers
from . import domain_logic, models, validators
from .jexl import QuestionJexl


class QuestionJexlField(serializers.JexlField):
    def __init__(self, **kwargs):
        super().__init__(QuestionJexl(), **kwargs)


class SaveFormSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Form
        fields = ["slug", "name", "description", "meta", "is_archived", "is_published"]


class CopyFormSerializer(serializers.ModelSerializer):
    source = serializers.GlobalIDPrimaryKeyRelatedField(
        queryset=models.Form.objects, required=True
    )

    @transaction.atomic
    def create(self, validated_data):
        return domain_logic.CopyFormLogic.copy(
            validated_data, user=self.context["request"].user
        )

    class Meta:
        model = models.Form
        fields = ["slug", "name", "description", "source", "is_published"]


class AddFormQuestionSerializer(serializers.ModelSerializer):
    form = serializers.GlobalIDField(source="slug")
    question = serializers.GlobalIDPrimaryKeyRelatedField(
        queryset=models.Question.objects
    )

    def update(self, instance, validated_data):
        # default sort is 0, as per default form question are sorted
        # in descending order this will be at the end
        _, created = models.FormQuestion.objects.get_or_create(
            form=self.instance, question=validated_data["question"]
        )

        if created:
            # reassign sort from start 1 so a newly added item with sort 0 will
            # be at the end again
            for sort, question in enumerate(
                self.instance.questions.all().order_by("formquestion__sort"), start=1
            ):
                models.FormQuestion.objects.filter(
                    form=instance, question=question
                ).update(sort=sort)

        return instance

    class Meta:
        fields = ["form", "question"]
        model = models.Form


class RemoveFormQuestionSerializer(serializers.ModelSerializer):
    form = serializers.GlobalIDField(source="slug")
    question = serializers.GlobalIDPrimaryKeyRelatedField(
        queryset=models.Question.objects
    )

    def update(self, instance, validated_data):
        models.FormQuestion.objects.filter(
            form=instance, question=validated_data["question"]
        ).delete()
        return instance

    class Meta:
        fields = ["form", "question"]
        model = models.Form


class FormQuestionRelatedField(serializers.GlobalIDPrimaryKeyRelatedField):
    def get_queryset(self):
        form = self.parent.parent.instance
        return form.questions.all()


class ReorderFormQuestionsSerializer(serializers.ModelSerializer):
    form = serializers.GlobalIDField(source="slug")
    questions = FormQuestionRelatedField(many=True)

    def update(self, instance, validated_data):
        questions = validated_data["questions"]
        curr_questions = set(instance.questions.all())

        if len(questions) != len(curr_questions) or set(questions) - set(
            instance.questions.all()
        ):
            raise exceptions.ValidationError(
                "Input questions are not the same as current form questions"
            )

        for sort, question in enumerate(reversed(questions), start=1):
            models.FormQuestion.objects.filter(form=instance, question=question).update(
                sort=sort
            )

        return instance

    class Meta:
        fields = ["form", "questions"]
        model = models.Form


class CopyQuestionSerializer(serializers.ModelSerializer):
    source = serializers.GlobalIDPrimaryKeyRelatedField(
        queryset=models.Question.objects, required=True
    )

    @transaction.atomic
    def create(self, validated_data):
        user = self.context["request"].user
        source = validated_data["source"]
        validated_data["type"] = source.type
        validated_data["is_required"] = source.is_required
        validated_data["is_hidden"] = source.is_hidden
        validated_data["configuration"] = dict(source.configuration)
        validated_data["meta"] = dict(source.meta)
        validated_data["row_form"] = source.row_form

        question = super().create(validated_data)

        new_question_options = [
            models.QuestionOption(
                sort=sort,
                question=question,
                option=question_option.option,
                created_by_user=user.username,
                created_by_group=user.group,
                modified_by_user=user.username,
                modified_by_group=user.group,
            )
            for sort, question_option in enumerate(
                reversed(models.QuestionOption.objects.filter(question=source)), start=1
            )
        ]
        for question_option in new_question_options:
            question_option.save()

        return question

    class Meta:
        model = models.Question
        fields = ["slug", "label", "source"]


class SaveQuestionSerializer(serializers.ModelSerializer):
    is_hidden = QuestionJexlField(required=False)
    is_required = QuestionJexlField(required=False)

    def validate(self, data):
        validators.QuestionValidator().validate(data)
        return super().validate(data)

    class Meta:
        model = models.Question
        fields = [
            "slug",
            "label",
            "info_text",
            "is_required",
            "is_hidden",
            "meta",
            "is_archived",
        ]


class SaveTextQuestionSerializer(SaveQuestionSerializer):
    min_length = IntegerField(min_value=1, required=False, allow_null=True)
    max_length = IntegerField(min_value=1, required=False, allow_null=True)
    format_validators = ListField(child=CharField(), required=False)

    def validate(self, data):
        data["type"] = models.Question.TYPE_TEXT
        return super().validate(data)

    class Meta(SaveQuestionSerializer.Meta):
        fields = SaveQuestionSerializer.Meta.fields + [
            "min_length",
            "max_length",
            "placeholder",
            "format_validators",
        ]


class SaveTextareaQuestionSerializer(SaveQuestionSerializer):
    min_length = IntegerField(min_value=1, required=False, allow_null=True)
    max_length = IntegerField(min_value=1, required=False, allow_null=True)
    format_validators = ListField(child=CharField(), required=False)

    def validate(self, data):
        data["type"] = models.Question.TYPE_TEXTAREA
        return super().validate(data)

    class Meta(SaveQuestionSerializer.Meta):
        fields = SaveQuestionSerializer.Meta.fields + [
            "min_length",
            "max_length",
            "placeholder",
            "format_validators",
        ]


class SaveDateQuestionSerializer(SaveQuestionSerializer):
    def validate(self, data):
        data["type"] = models.Question.TYPE_DATE
        return super().validate(data)

    class Meta(SaveQuestionSerializer.Meta):
        fields = SaveQuestionSerializer.Meta.fields


class SaveQuestionOptionsMixin(object):
    def create_question_options(self, question, options):
        user = self.context["request"].user
        question_options = [
            models.QuestionOption(
                sort=sort,
                question=question,
                option=option,
                created_by_user=user.username,
                created_by_group=user.group,
                modified_by_user=user.username,
                modified_by_group=user.group,
            )
            for sort, option in enumerate(reversed(options), start=1)
        ]
        for question_option in question_options:
            question_option.save()

    @transaction.atomic
    def create(self, validated_data):
        options = validated_data.pop("options")
        instance = super().create(validated_data)
        self.create_question_options(instance, options)
        return instance

    @transaction.atomic
    def update(self, instance, validated_data):
        options = validated_data.pop("options")
        models.QuestionOption.objects.filter(question=instance).delete()
        instance = super().update(instance, validated_data)
        self.create_question_options(instance, options)
        return instance


class SaveMultipleChoiceQuestionSerializer(
    SaveQuestionOptionsMixin, SaveQuestionSerializer
):
    options = serializers.GlobalIDPrimaryKeyRelatedField(
        queryset=models.Option.objects.all(), many=True, required=True
    )

    def validate(self, data):
        data["type"] = models.Question.TYPE_MULTIPLE_CHOICE
        return super().validate(data)

    class Meta(SaveQuestionSerializer.Meta):
        fields = SaveQuestionSerializer.Meta.fields + ["options"]


class SaveChoiceQuestionSerializer(SaveQuestionOptionsMixin, SaveQuestionSerializer):
    options = serializers.GlobalIDPrimaryKeyRelatedField(
        queryset=models.Option.objects.all(), many=True, required=True
    )

    def validate(self, data):
        data["type"] = models.Question.TYPE_CHOICE
        return super().validate(data)

    class Meta(SaveQuestionSerializer.Meta):
        fields = SaveQuestionSerializer.Meta.fields + ["options"]


class SaveDynamicChoiceQuestionSerializer(SaveQuestionSerializer):
    data_source = CharField()

    def validate(self, data):
        data["type"] = models.Question.TYPE_DYNAMIC_CHOICE
        return super().validate(data)

    class Meta(SaveQuestionSerializer.Meta):
        fields = SaveQuestionSerializer.Meta.fields + ["data_source"]


class SaveDynamicMultipleChoiceQuestionSerializer(SaveQuestionSerializer):
    data_source = CharField()

    def validate(self, data):
        data["type"] = models.Question.TYPE_DYNAMIC_MULTIPLE_CHOICE
        return super().validate(data)

    class Meta(SaveQuestionSerializer.Meta):
        fields = SaveQuestionSerializer.Meta.fields + ["data_source"]


class SaveFloatQuestionSerializer(SaveQuestionSerializer):
    min_value = FloatField(required=False, allow_null=True)
    max_value = FloatField(required=False, allow_null=True)

    def validate(self, data):
        min_value = (
            data.get("min_value")
            if data.get("min_value") is not None
            else float("-inf")
        )
        max_value = (
            data.get("max_value") if data.get("max_value") is not None else float("inf")
        )

        if min_value > max_value:
            raise exceptions.ValidationError(
                f"max_value {max_value} is smaller than {min_value}"
            )

        data["type"] = models.Question.TYPE_FLOAT
        return super().validate(data)

    class Meta(SaveQuestionSerializer.Meta):
        fields = SaveQuestionSerializer.Meta.fields + [
            "min_value",
            "max_value",
            "placeholder",
        ]


class SaveIntegerQuestionSerializer(SaveQuestionSerializer):
    min_value = IntegerField(required=False, allow_null=True)
    max_value = IntegerField(required=False, allow_null=True)

    def validate(self, data):
        min_value = (
            data.get("min_value")
            if data.get("min_value") is not None
            else float("-inf")
        )
        max_value = (
            data.get("max_value") if data.get("max_value") is not None else float("inf")
        )

        if min_value > max_value:
            raise exceptions.ValidationError(
                f"max_value {max_value} is smaller than {min_value}"
            )

        data["type"] = models.Question.TYPE_INTEGER
        return super().validate(data)

    class Meta(SaveQuestionSerializer.Meta):
        fields = SaveQuestionSerializer.Meta.fields + [
            "min_value",
            "max_value",
            "placeholder",
        ]


class SaveTableQuestionSerializer(SaveQuestionSerializer):
    row_form = serializers.GlobalIDPrimaryKeyRelatedField(
        queryset=models.Form.objects,
        required=True,
        help_text=models.Question._meta.get_field("row_form").help_text,
    )

    def validate(self, data):
        data["type"] = models.Question.TYPE_TABLE
        return super().validate(data)

    class Meta(SaveQuestionSerializer.Meta):
        fields = SaveQuestionSerializer.Meta.fields + ["row_form"]


class SaveFormQuestionSerializer(SaveQuestionSerializer):
    sub_form = serializers.GlobalIDPrimaryKeyRelatedField(
        queryset=models.Form.objects, required=True
    )

    def validate(self, data):
        data["type"] = models.Question.TYPE_FORM
        return super().validate(data)

    class Meta(SaveQuestionSerializer.Meta):
        fields = SaveQuestionSerializer.Meta.fields + ["sub_form"]


class SaveFileQuestionSerializer(SaveQuestionSerializer):
    def validate(self, data):
        data["type"] = models.Question.TYPE_FILE
        return super().validate(data)

    class Meta(SaveQuestionSerializer.Meta):
        fields = SaveQuestionSerializer.Meta.fields


class SaveStaticQuestionSerializer(SaveQuestionSerializer):
    static_content = CharField(required=False)

    def validate(self, data):
        data["type"] = models.Question.TYPE_STATIC
        return super().validate(data)

    class Meta(SaveQuestionSerializer.Meta):
        fields = [
            "label",
            "slug",
            "info_text",
            "is_hidden",
            "meta",
            "is_archived",
            "static_content",
        ]


class SaveCalculatedFloatQuestionSerializer(SaveQuestionSerializer):
    calc_expression = QuestionJexlField(required=False)

    def validate(self, data):
        data["type"] = models.Question.TYPE_CALCULATED_FLOAT
        return super().validate(data)

    class Meta(SaveQuestionSerializer.Meta):
        fields = SaveQuestionSerializer.Meta.fields + [
            "calc_expression",
        ]


class SaveOptionSerializer(serializers.ModelSerializer):
    class Meta:
        fields = ["slug", "label", "is_archived", "meta"]
        model = models.Option


class CopyOptionSerializer(serializers.ModelSerializer):
    source = serializers.GlobalIDPrimaryKeyRelatedField(
        queryset=models.Option.objects, required=True
    )

    def create(self, validated_data):
        source = validated_data["source"]
        validated_data["meta"] = dict(source.meta)
        return super().create(validated_data)

    class Meta:
        fields = ["slug", "label", "source"]
        model = models.Option


class DocumentSerializer(serializers.ModelSerializer):
    def create(self, validated_data):
        return domain_logic.SaveDocumentLogic.create(
            validated_data, user=self.context["request"].user
        )

    class Meta:
        model = models.Document
        fields = [
            "id",
            "form",
            "meta",
        ]


class SaveAnswerSerializer(serializers.ModelSerializer):
    def validate(self, data):
        data = domain_logic.SaveAnswerLogic.validate_for_save(
            data, self.context["request"].user, self.instance, True
        )
        return super().validate(data)

    @transaction.atomic
    def create(self, validated_data):
        return domain_logic.SaveAnswerLogic.create(
            validated_data, user=self.context["request"].user
        )

    @transaction.atomic
    def update(self, instance, validated_data):
        return domain_logic.SaveAnswerLogic.update(instance, validated_data)

    class Meta:
        model = models.Answer
        fields = ["question", "document", "meta", "value"]


class SaveDocumentStringAnswerSerializer(SaveAnswerSerializer):
    value = CharField(trim_whitespace=False, allow_blank=True, required=False)

    class Meta(SaveAnswerSerializer.Meta):
        pass


class SaveDocumentListAnswerSerializer(SaveAnswerSerializer):
    value = ListField(child=CharField(), required=False)

    class Meta(SaveAnswerSerializer.Meta):
        pass


class SaveDocumentIntegerAnswerSerializer(SaveAnswerSerializer):
    value = IntegerField(required=False)

    class Meta(SaveAnswerSerializer.Meta):
        pass


class SaveDocumentFloatAnswerSerializer(SaveAnswerSerializer):
    value = FloatField(required=False)

    class Meta(SaveAnswerSerializer.Meta):
        pass


class SaveDocumentDateAnswerSerializer(SaveAnswerSerializer):
    value = DateField(source="date", required=False)

    class Meta(SaveAnswerSerializer.Meta):
        pass


class SaveDocumentTableAnswerSerializer(SaveAnswerSerializer):
    value = serializers.GlobalIDPrimaryKeyRelatedField(
        source="documents",
        queryset=models.Document.objects,
        many=True,
        required=False,
        help_text="List of document IDs representing the rows in the table.",
    )

    class Meta(SaveAnswerSerializer.Meta):
        pass


class SaveDocumentFileAnswerSerializer(SaveAnswerSerializer):
    value = CharField(write_only=True, source="file", required=False)
    value_id = PrimaryKeyRelatedField(read_only=True, source="file", required=False)

    class Meta(SaveAnswerSerializer.Meta):
        fields = SaveAnswerSerializer.Meta.fields + [
            "value_id",
        ]


class RemoveAnswerSerializer(serializers.ModelSerializer):
    answer = PrimaryKeyRelatedField(
        queryset=models.Answer.objects.filter(document__isnull=False)
    )

    @transaction.atomic
    def update(self, instance, validated_data):
        instance.delete()
        return instance

    class Meta:
        fields = [
            "answer",
        ]
        model = models.Answer


class SaveDefaultAnswerSerializer(serializers.ModelSerializer):
    def validate(self, data):
        return domain_logic.SaveDefaultAnswerLogic.validate_for_save(
            data, self.context["request"].user, self.instance, True
        )

    def create(self, validated_data):
        return domain_logic.SaveDefaultAnswerLogic.create(
            validated_data, self.context["request"].user
        )

    def update(self, instance, validated_data):
        return domain_logic.SaveDefaultAnswerLogic.update(instance, validated_data)

    class Meta:
        model = models.Answer
        fields = ["question", "meta", "value"]


class SaveDefaultStringAnswerSerializer(SaveDefaultAnswerSerializer):
    value = CharField(trim_whitespace=False, allow_blank=True, required=False)

    class Meta(SaveDefaultAnswerSerializer.Meta):
        pass


class SaveDefaultListAnswerSerializer(SaveDefaultAnswerSerializer):
    value = ListField(child=CharField(), required=False)

    class Meta(SaveDefaultAnswerSerializer.Meta):
        pass


class SaveDefaultIntegerAnswerSerializer(SaveDefaultAnswerSerializer):
    value = IntegerField(required=False)

    class Meta(SaveDefaultAnswerSerializer.Meta):
        pass


class SaveDefaultFloatAnswerSerializer(SaveDefaultAnswerSerializer):
    value = FloatField(required=False)

    class Meta(SaveDefaultAnswerSerializer.Meta):
        pass


class SaveDefaultDateAnswerSerializer(SaveDefaultAnswerSerializer):
    value = DateField(source="date", required=False)

    class Meta(SaveDefaultAnswerSerializer.Meta):
        pass


class SaveDefaultTableAnswerSerializer(SaveDefaultAnswerSerializer):
    value = serializers.GlobalIDPrimaryKeyRelatedField(
        source="documents",
        queryset=models.Document.objects,
        many=True,
        required=False,
        help_text="List of document IDs representing the rows in the table.",
    )

    class Meta(SaveDefaultAnswerSerializer.Meta):
        pass


class RemoveDefaultAnswerSerializer(serializers.ModelSerializer):
    question = PrimaryKeyRelatedField(queryset=models.Question.objects)

    @transaction.atomic
    def update(self, instance, validated_data):
        instance.default_answer.delete()
        return instance

    class Meta:
        fields = [
            "question",
        ]
        model = models.Question


class RemoveDocumentSerializer(serializers.ModelSerializer):
    document = serializers.GlobalIDField(source="id")

    @transaction.atomic
    def update(self, instance, validated_data):
        if hasattr(instance, "case"):
            raise Exception("You cannot remove a Document, if it's attached to a case.")

        for answer in instance.answers.filter(
            question__type=models.Question.TYPE_TABLE
        ):
            answer.documents.all().delete()

        instance.delete()

        return instance

    class Meta:
        fields = ["document"]
        model = models.Document


class CopyDocumentSerializer(serializers.ModelSerializer):
    source = serializers.GlobalIDPrimaryKeyRelatedField(
        queryset=models.Document.objects, required=True
    )

    @transaction.atomic
    def create(self, validated_data):
        return validated_data["source"].copy(
            family=None,
            user=self.context["request"].user,
        )

    class Meta:
        model = models.Document
        fields = ["source"]



import faker
from factory import Faker, LazyAttribute, Maybe, SubFactory, lazy_attribute

from ..caluma_core.factories import DjangoModelFactory
from . import models

AUTO_QUESTION_TYPES = [
    t
    for t in models.Question.TYPE_CHOICES
    if t
    not in [
        models.Question.TYPE_STATIC,
        models.Question.TYPE_FORM,
        models.Question.TYPE_DYNAMIC_CHOICE,
        models.Question.TYPE_DYNAMIC_MULTIPLE_CHOICE,
        models.Question.TYPE_CALCULATED_FLOAT,
    ]
]


class FormFactory(DjangoModelFactory):
    slug = Faker("slug")
    name = Faker("multilang", faker_provider="name")
    description = Faker("multilang", faker_provider="text")
    meta = {}
    is_published = False
    is_archived = False

    class Meta:
        model = models.Form


class QuestionFactory(DjangoModelFactory):
    slug = Faker("slug")
    label = Faker("multilang", faker_provider="name")
    type = Faker("word", ext_word_list=AUTO_QUESTION_TYPES)
    is_required = "true"
    is_hidden = "false"
    configuration = {}
    meta = {}
    is_archived = False
    format_validators = []

    row_form = Maybe(
        "is_table", yes_declaration=SubFactory(FormFactory), no_declaration=None
    )
    sub_form = Maybe(
        "is_form", yes_declaration=SubFactory(FormFactory), no_declaration=None
    )
    static_content = Maybe(
        "is_static",
        yes_declaration=Faker("multilang", faker_provider="text"),
        no_declaration=None,
    )

    data_source = Maybe(
        "is_dynamic", yes_declaration="MyDataSource", no_declaration=None
    )
    calc_expression = Maybe(
        "is_calc_float", yes_declaration="-1.0", no_declaration=None
    )
    calc_dependents = []

    class Meta:
        model = models.Question

    class Params:
        is_table = LazyAttribute(lambda q: q.type == models.Question.TYPE_TABLE)
        is_form = LazyAttribute(lambda q: q.type == models.Question.TYPE_FORM)
        is_dynamic = LazyAttribute(
            lambda q: q.type
            in [
                models.Question.TYPE_DYNAMIC_CHOICE,
                models.Question.TYPE_DYNAMIC_MULTIPLE_CHOICE,
            ]
        )
        is_static = LazyAttribute(lambda q: q.type == models.Question.TYPE_STATIC)
        is_calc_float = LazyAttribute(
            lambda q: q.type == models.Question.TYPE_CALCULATED_FLOAT
        )


class OptionFactory(DjangoModelFactory):
    slug = Faker("slug")
    label = Faker("multilang", faker_provider="name")
    is_archived = False
    meta = {}

    class Meta:
        model = models.Option


class QuestionOptionFactory(DjangoModelFactory):
    option = SubFactory(OptionFactory)
    question = SubFactory(QuestionFactory)
    sort = 0

    class Meta:
        model = models.QuestionOption


class FormQuestionFactory(DjangoModelFactory):
    form = SubFactory(FormFactory)
    question = SubFactory(QuestionFactory)
    sort = 0

    class Meta:
        model = models.FormQuestion


class DocumentFactory(DjangoModelFactory):
    form = SubFactory(FormFactory)
    family = None
    meta = {}

    class Meta:
        model = models.Document


class FileFactory(DjangoModelFactory):
    name = Faker("file_name")

    class Meta:
        model = models.File


class AnswerFactory(DjangoModelFactory):
    question = SubFactory(QuestionFactory)
    document = SubFactory(DocumentFactory)
    meta = {}

    @lazy_attribute
    def value(self):
        if (
            self.question.type == models.Question.TYPE_MULTIPLE_CHOICE
            or self.question.type == models.Question.TYPE_DYNAMIC_MULTIPLE_CHOICE
        ):
            return [faker.Faker().name(), faker.Faker().name()]
        elif self.question.type == models.Question.TYPE_FLOAT:
            return faker.Faker().pyfloat()
        elif self.question.type == models.Question.TYPE_INTEGER:
            return faker.Faker().pyint()
        elif self.question.type not in [
            models.Question.TYPE_TABLE,
            models.Question.TYPE_FILE,
            models.Question.TYPE_DATE,
        ]:
            return faker.Faker().name()

        return None

    file = Maybe(
        "is_file", yes_declaration=SubFactory(FileFactory), no_declaration=None
    )
    date = Maybe("is_date", yes_declaration=Faker("date"), no_declaration=None)

    class Meta:
        model = models.Answer

    class Params:
        is_file = LazyAttribute(lambda a: a.question.type == models.Question.TYPE_FILE)
        is_date = LazyAttribute(lambda a: a.question.type == models.Question.TYPE_DATE)


class AnswerDocumentFactory(DjangoModelFactory):
    answer = SubFactory(AnswerFactory)
    document = SubFactory(DocumentFactory)
    sort = 0

    class Meta:
        model = models.AnswerDocument


class DynamicOptionFactory(DjangoModelFactory):
    slug = Faker("slug")
    label = Faker("multilang", faker_provider="name")
    document = SubFactory(DocumentFactory)
    question = SubFactory(QuestionFactory)

    class Meta:
        model = models.DynamicOption



from functools import wraps


def register_dynamic_group(dynamic_group_name):
    """Decorate function to register a dynamic group."""

    def wrapper(fn, *arg, **kwarg):
        fn._dynamic_group_name = dynamic_group_name

        @wraps(fn)
        def wrapped(self, *args, **kwargs):
            return fn(self, *args, **kwargs)

        return wrapped

    return wrapper


class BaseDynamicGroups:
    """Basic dynamic groups class to be extended by any dynamic groups class implementation.

    In combination with the decorator `@register_dynamic_group` a dynamic group
    class can extend the JEXL `groups` transform used in the `address_groups`
    and `control_groups` property of a task.

    A dynamic groups class could look like this:

    >>> from your.project.code import find_legal_dept
    ... from caluma.caluma_workflow.dynamic_groups import (
    ...     BaseDynamicGroups,
    ...     register_dynamic_group,
    ... )
    ...
    ...
    ... class CustomDynamicGroups(BaseDynamicGroups):
    ...     @register_dynamic_group("legal_dept")
    ...     def resolve_legal_dept(self, task, case, user, prev_work_item, context):
    ...         # custom business logic which returns either a string or a list of strings
    ...         return find_legal_dept(case)
    """

    def resolve(self, dynamic_group_name):
        for methodname in dir(self):
            method = getattr(self, methodname)
            if (
                hasattr(method, "_dynamic_group_name")
                and method._dynamic_group_name == dynamic_group_name
            ):
                return method



# -*- coding: utf-8 -*-
# Generated by Django 1.11.17 on 2018-12-13 13:34
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_workflow", "0001_initial")]

    operations = [
        migrations.AlterField(
            model_name="case",
            name="document",
            field=models.OneToOneField(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="case",
                to="caluma_form.Document",
            ),
        )
    ]



from functools import partial
from itertools import chain

from django.core.exceptions import ValidationError
from pyjexl.analysis import ValidatingAnalyzer
from pyjexl.evaluator import Context
from pyjexl.parser import Literal

from ..caluma_core.jexl import JEXL, ExtractTransformSubjectAnalyzer
from .models import Task


def task_exists(slug):
    return (
        slug in FlowJexl.get_all_registered_dynamic_tasks()
        or Task.objects.filter(pk=slug).exists()
    )


def parse_literal(value):
    if isinstance(value, Literal):
        return value.value

    if isinstance(value, list):
        return [parse_literal(item) for item in value]

    return value


class GroupValidatingAnalyzer(ValidatingAnalyzer):
    def visit_Transform(self, transform):
        if transform.name == "groups" and not isinstance(transform.subject.value, list):
            yield f"{transform.subject.value} is not a valid list of groups."

        yield from super().visit_Transform(transform)


class TaskValidatingAnalyzer(ValidatingAnalyzer):
    def visit_Transform(self, transform):
        value = parse_literal(transform.subject.value)

        if transform.name == "tasks":
            if not isinstance(value, list):
                yield f"`{value}` is not a valid list of tasks."
            else:
                for slug in value:
                    if not task_exists(slug):
                        yield f"The task `{slug}` does not exist or is not registered as dynamic task"

        elif transform.name == "task":
            if not isinstance(value, str):
                yield f"`{value}` is not a valid task slug."
            elif not task_exists(value):
                yield f"The task `{value}` does not exist or is not registered as dynamic task"

        yield from super().visit_Transform(transform)


class GroupJexl(JEXL):
    """
    Class for evaluating GroupJexl.

    validation_context is expected to be the following:

    {
        "case": {
            "created_by_group": str,
        },
        "work_item": {
            "created_by_group": str,
        },
        "prev_work_item": {
            "controlling_groups": list of str,
        },
    }
    """

    def __init__(
        self,
        validation_context=None,
        task=None,
        case=None,
        work_item_created_by_user=None,
        prev_work_item=None,
        dynamic_context=None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        context_data = None

        if validation_context:
            context_data = {"info": validation_context}

        self.context = Context(context_data)

        self.task = task
        self.case = case
        self.work_item_created_by_user = work_item_created_by_user
        self.prev_work_item = prev_work_item
        self.dynamic_context = dynamic_context

        self.add_transform("groups", self.groups_transform)

    def validate(self, expression):
        return super().validate(expression, GroupValidatingAnalyzer)

    def evaluate(self, expression, context=None):
        value = super().evaluate(expression, context)
        if isinstance(value, list) or value is None:
            return value
        return [value]

    def groups_transform(self, group_names):
        evaluated = [self._get_dynamic_group(group_name) for group_name in group_names]

        return sorted(set(chain(*evaluated)))

    def _get_dynamic_group(self, name):
        for dynamic_groups_class in self.dynamic_groups_classes:
            method = dynamic_groups_class().resolve(name)
            if method:
                value = method(
                    self.task,
                    self.case,
                    self.work_item_created_by_user,
                    self.prev_work_item,
                    self.dynamic_context,
                )

                if not isinstance(value, list):
                    value = [value]

                return value

        return [name]


class FlowJexl(JEXL):
    def __init__(
        self, case=None, user=None, prev_work_item=None, dynamic_context=None, **kwargs
    ):
        super().__init__(**kwargs)

        self.case = case
        self.user = user
        self.prev_work_item = prev_work_item
        self.dynamic_context = dynamic_context

        self.add_transform("task", self.task_transform)
        self.add_transform("tasks", self.tasks_transform)

    def validate(self, expression):
        return super().validate(expression, TaskValidatingAnalyzer)

    def extract_tasks(self, expr):
        yield from self.analyze(
            expr, partial(ExtractTransformSubjectAnalyzer, transforms=["task"])
        )

        # tasks transforms return a list of literals
        yield from (
            literal.value
            for literal in chain(
                *self.analyze(
                    expr, partial(ExtractTransformSubjectAnalyzer, transforms=["tasks"])
                )
            )
        )

    @classmethod
    def get_all_registered_dynamic_tasks(cls):
        return chain(
            *[
                dynamic_tasks_class().get_registered_dynamic_tasks()
                for dynamic_tasks_class in cls.dynamic_tasks_classes
            ]
        )

    def task_transform(self, task_name):
        all_tasks = self.tasks_transform([task_name])

        if len(all_tasks) > 1:
            raise ValidationError(
                f"The dynamic task `{task_name}` used by the `task` transform should "
                f"not return more than one task -- it returned {len(all_tasks)}"
            )

        return all_tasks[0] if len(all_tasks) else []

    def tasks_transform(self, task_names):
        evaluated = [self._get_dynamic_task(task_name) for task_name in task_names]

        # transform all evaluated tasks into a sorted, unique and flat list
        return sorted(set(chain(*evaluated)))

    def _get_dynamic_task(self, name):
        for dynamic_tasks_class in self.dynamic_tasks_classes:
            method = dynamic_tasks_class().resolve(name)

            if method:
                value = method(
                    self.case, self.user, self.prev_work_item, self.dynamic_context
                )

                if not isinstance(value, list):
                    value = [value]

                return value

        return [name]



from ...caluma_core.relay import extract_global_id
from .. import models


def test_root_document_filter(
    schema_executor,
    db,
    form_factory,
    form_question_factory,
    document_factory,
    answer_factory,
):
    # build up our form first
    top_form = form_factory()
    top_question = form_question_factory.create(
        form=top_form,
        question__slug="top_question",
        question__type=models.Question.TYPE_TEXT,
    ).question
    subform_question = form_question_factory.create(
        form=top_form,
        question__slug="subform-question",
        question__type=models.Question.TYPE_FORM,
        question__sub_form__slug="subform",
    ).question

    sub_questions = [
        form_question_factory(
            question__slug="foo",
            question__type=models.Question.TYPE_TEXT,
            form=subform_question.sub_form,
        ).question,
        form_question_factory(
            question__slug="bar",
            question__type=models.Question.TYPE_TEXT,
            form=subform_question.sub_form,
        ).question,
    ]

    document1 = document_factory(form=top_form)
    document2 = document_factory(form=top_form, family=document1)
    document3 = document_factory(form=top_form)

    for d in [document1, document2]:
        for question in sub_questions:
            answer_factory(question=question, document=d, value="hello")

    answer_factory(question=top_question, document=document3, value="foo")

    # fetch documents, see if subdocuments are included
    query = """
        query asdf ($root: ID!) {
          allDocuments(rootDocument: $root) {
            edges {
              node {
                id
              }
            }
          }
        }
    """
    variables = {"root": str(document1.id)}

    result = schema_executor(query, variable_values=variables)
    assert not result.errors
    result_ids = [
        extract_global_id(doc["node"]["id"])
        for doc in result.data["allDocuments"]["edges"]
    ]

    assert sorted(result_ids) == sorted([str(document1.id), str(document2.id)])



import os

import pytest
from django.core.management import call_command
from django.db import connection


@pytest.mark.parametrize("force", [True, False])
def test_migrate_to_prefixed_apps(db, force):
    def _is_applied(query):
        with connection.cursor() as cursor:
            cursor.execute(query)
            if cursor.fetchone():
                return True
        return False

    def changes_applied():
        applied = (
            _is_applied(
                """SELECT "app_label" FROM "django_content_type" WHERE "app_label" = 'caluma_form';"""
            ),
            _is_applied(
                """SELECT "app" FROM "django_migrations" WHERE "app" = 'caluma_form';"""
            ),
            _is_applied(
                """SELECT tablename FROM pg_catalog.pg_tables WHERE tablename LIKE 'caluma_form_%' OR tablename LIKE 'caluma_workflow_%';"""
            ),
            _is_applied(
                """SELECT sequence_name FROM information_schema.sequences WHERE sequence_name LIKE 'caluma_form_%' OR sequence_name LIKE 'caluma_workflow_%';"""
            ),
        )
        if all(applied):
            return True
        elif not any(applied):
            return False
        else:  # pragma: no cover
            raise Exception("DB is in a state it should never be!")

    kwargs = {}
    if force:
        kwargs["force"] = force

    assert changes_applied()

    call_command(
        "migrate_to_prefixed_apps", **kwargs, revert=True, stderr=open(os.devnull, "w")
    )

    if force:
        assert not changes_applied()
    else:
        assert changes_applied()

    call_command("migrate_to_prefixed_apps", **kwargs, stderr=open(os.devnull, "w"))

    assert changes_applied()



def test_search_in_filter_collection(db, schema_executor, form_factory):
    form_factory(name="Test 1")
    form_factory(name="Test 2")
    form_factory(name="Test 3")

    query = """
        query($search: String!) {
          allForms(filter: [{ search: $search }]) {
            edges {
              node {
                slug
                name
              }
            }
          }
        }
    """

    result = schema_executor(query, variable_values={"search": "Test 2"})
    assert not result.errors

    assert len(result.data["allForms"]["edges"]) == 1
    assert result.data["allForms"]["edges"][0]["node"]["name"] == "Test 2"



# Generated by Django 2.2.17 on 2021-01-14 09:56

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("caluma_form", "0038_auto_20201224_0920"),
        ("caluma_form", "0038_remove_table_answer_value"),
    ]

    operations = []



# Generated by Django 2.2.17 on 2020-11-23 14:28

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0036_default_answers")]

    operations = [
        migrations.AlterField(
            model_name="question",
            name="default_answer",
            field=models.OneToOneField(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="caluma_form.Answer",
            ),
        )
    ]



import uuid

from django.contrib.postgres.operations import HStoreExtension
from django.db import connection, migrations, models
from django.db.migrations.executor import MigrationExecutor


def define_fake_model(fields=None, model_base=models.Model, options=None):
    fields = fields if fields else {}
    options = options if options else {}
    name = str(uuid.uuid4()).replace("-", "")[:8]

    meta_options = {"app_label": "caluma_core"}
    meta_options.update(options)

    attributes = {
        "__module__": __name__,
        "__name__": name,
        "Meta": type("Meta", (object,), meta_options),
    }

    attributes.update(fields)
    model = type(name, (model_base,), attributes)

    return model


def get_fake_model(fields=None, model_base=models.Model, options=None):
    fields = fields if fields else {}
    options = options if options else {}
    """Create fake model to use during unit tests."""

    model = define_fake_model(fields, model_base, options)

    class TestProject:
        def clone(self, *_args, **_kwargs):
            return self

    class TestMigration(migrations.Migration):
        operations = [HStoreExtension()]

    with connection.schema_editor() as schema_editor:
        migration_executor = MigrationExecutor(schema_editor.connection)
        migration_executor.apply_migration(
            TestProject(), TestMigration("caluma_extra", "caluma_core")
        )

        schema_editor.create_model(model)

    return model



import pytest
from django.core.exceptions import ValidationError

from caluma.caluma_workflow.dynamic_tasks import BaseDynamicTasks, register_dynamic_task
from caluma.caluma_workflow.jexl import FlowJexl


class CustomDynamicTasks(BaseDynamicTasks):
    @register_dynamic_task("task-1")
    def resolve_task_1(self, case, user, prev_work_item, context):
        return "some-task"

    @register_dynamic_task("task-2")
    def resolve_task_2(self, case, user, prev_work_item, context):
        return ["some-task", "some-other-task"]


@pytest.mark.parametrize(
    "expression,success,expected_result",
    [
        ('"task-1"|task', True, "some-task"),
        ('["task-1"]|tasks', True, ["some-task"]),
        ('["task-2"]|tasks', True, ["some-task", "some-other-task"]),
        ('["task-1", "task-2"]|tasks', True, ["some-task", "some-other-task"]),
        # task transform can only return single values
        ('"task-2"|task', False, None),
    ],
)
def test_dynamic_tasks(
    db, case, admin_user, work_item, expression, success, expected_result
):
    FlowJexl.dynamic_tasks_classes = [CustomDynamicTasks]

    jexl = FlowJexl(case, admin_user, work_item, {})

    if success:
        assert sorted(jexl.evaluate(expression)) == sorted(expected_result)
    else:
        with pytest.raises(ValidationError) as error:
            jexl.evaluate(expression)

        assert (
            "The dynamic task `task-2` used by the `task` transform should not return more than one task -- it returned 2"
            in str(error.value)
        )



import pytest
from graphql_relay import to_global_id

from .. import models


@pytest.mark.parametrize(
    "question__type,expected_typename,success",
    [
        (models.Question.TYPE_FILE, "FileAnswer", True),
        (models.Question.TYPE_TEXT, "StringAnswer", True),
        (models.Question.TYPE_TEXTAREA, "StringAnswer", True),
        (models.Question.TYPE_FLOAT, "FloatAnswer", True),
        (models.Question.TYPE_CHOICE, "StringAnswer", True),
        (models.Question.TYPE_INTEGER, "IntegerAnswer", True),
        (models.Question.TYPE_MULTIPLE_CHOICE, "ListAnswer", True),
        (models.Question.TYPE_DYNAMIC_CHOICE, "StringAnswer", True),
        (models.Question.TYPE_DYNAMIC_MULTIPLE_CHOICE, "ListAnswer", True),
        (models.Question.TYPE_DATE, "DateAnswer", True),
        (models.Question.TYPE_TABLE, "TableAnswer", True),
        (models.Question.TYPE_STATIC, "StringAnswer", False),
    ],
)
def test_answer_types(
    db, question, expected_typename, success, answer_factory, schema_executor
):
    answer = answer_factory(question=question)
    global_id = to_global_id(expected_typename, answer.pk)

    query = """
        query Answer($id: ID!) {
            node(id: $id) {
              id
              __typename
            }
        }
    """

    result = schema_executor(query, variable_values={"id": global_id})
    assert not result.errors == success
    if success:
        assert result.data["node"]["__typename"] == expected_typename


@pytest.mark.parametrize(
    "question__type,expected_typename",
    [
        (models.Question.TYPE_TEXT, "TextQuestion"),
        (models.Question.TYPE_FLOAT, "FloatQuestion"),
        (models.Question.TYPE_CHOICE, "ChoiceQuestion"),
        (models.Question.TYPE_INTEGER, "IntegerQuestion"),
        (models.Question.TYPE_MULTIPLE_CHOICE, "MultipleChoiceQuestion"),
        (models.Question.TYPE_DYNAMIC_CHOICE, "DynamicChoiceQuestion"),
        (models.Question.TYPE_DYNAMIC_MULTIPLE_CHOICE, "DynamicMultipleChoiceQuestion"),
        (models.Question.TYPE_TEXTAREA, "TextareaQuestion"),
        (models.Question.TYPE_DATE, "DateQuestion"),
        (models.Question.TYPE_TABLE, "TableQuestion"),
        (models.Question.TYPE_FORM, "FormQuestion"),
        (models.Question.TYPE_FILE, "FileQuestion"),
        (models.Question.TYPE_STATIC, "StaticQuestion"),
    ],
)
def test_question_types(
    db, question, expected_typename, answer_factory, schema_executor
):
    global_id = to_global_id(expected_typename, question.pk)

    query = """
        query Question($id: ID!) {
            node(id: $id) {
              id
              __typename
            }
        }
    """

    result = schema_executor(query, variable_values={"id": global_id})
    assert not result.errors
    assert result.data["node"]["__typename"] == expected_typename



import pytest

from .. import models, structure, validators
from ..jexl import QuestionJexl, QuestionMissing
from ..models import Question


@pytest.mark.parametrize(
    "expression,num_errors",
    [
        # correct case
        ("\"question-slug\"|answer|mapby('key')", 0),
        # invalid subject type
        ("100|answer", 1),
        # two invalid subject types
        ('["test"]|answer || 1.0|answer', 2),
        # invalid operator
        ("'question-slug1'|answer &&& 'question-slug2'|answer", 1),
    ],
)
def test_question_jexl_validate(expression, num_errors):
    jexl = QuestionJexl()
    assert len(list(jexl.validate(expression))) == num_errors


@pytest.mark.parametrize(
    "expression,result",
    [
        ("[1,2] intersects [2,3]", True),
        ("[1,2] intersects [3,4]", False),
        ("[] intersects []", False),
        ("[1] intersects []", False),
        ("['foo'] intersects ['bar', 'bazz']", False),
        ("['foo'] intersects ['foo', 'foo']", True),
        ("[1] intersects [1] && [2] intersects [2]", True),
        ("[2] intersects [1] + [2]", True),
    ],
)
def test_intersects_operator(expression, result):
    assert QuestionJexl().evaluate(expression) == result


@pytest.mark.parametrize("form__slug", ["f-main-slug"])
def test_jexl_form(db, form):
    answer_by_question = {
        "a1": {"value": "A1", "form": form},
        "b1": {"value": "B1", "form": form},
    }

    assert (
        QuestionJexl({"answers": answer_by_question, "form": form}).evaluate("form")
        == "f-main-slug"
    )


def test_all_deps_hidden(db, form, document_factory, form_question_factory):
    q1 = form_question_factory(form=form, question__is_hidden="true").question
    q2 = form_question_factory(
        form=form,
        question__is_hidden=f"'{q1.slug}'|answer=='blah'",
        question__is_required=f"'{q1.slug}'|answer=='blah'",
    ).question
    document = document_factory(form=form)

    qj = QuestionJexl(
        {
            "document": document,
            "answers": {},
            "form": form,
            "structure": structure.FieldSet(document, document.form),
        }
    )
    field = structure.Field(document, document.form, q2)
    assert qj.is_hidden(field)
    assert not qj.is_required(field)


@pytest.mark.parametrize("fq_is_hidden", ["true", "false"])
def test_indirectly_hidden_dependency(
    db,
    form_question_factory,
    form_factory,
    question_factory,
    document_factory,
    answer_factory,
    info,
    fq_is_hidden,
):
    # Questions can not only be hidden by evaluating their is_hidden expression
    # but also if they're part of a subform, where the containing form question
    # is hidden.
    #
    # Showcase form to demonstrate this (f:form, q:question, d:document, a:answer):
    # f:topform
    #    q:formquestionbranch - hidden=false
    #      f:subformbranch
    #         q:depquestion2 - required = 'subquestion2'|answer=='bluh'
    #    q:formquestion - hidden=variable
    #      f:subform
    #         q:subquestion1 - hidden=false
    #         q:subquestion2 - hidden=false
    #   q:depquestion1 - required = 'subquestion1'|answer=='blah'
    #
    # d:topdoc f=topform
    #    a:subanswer1 q=subquestion1, value='blah'
    #    a:subanswer2 q=subquestion2, value='bluh'
    #    (depquestion1 and depquestion2 have no answers, as they're indirectly not required)
    #
    # Note: The above structure shows the actual "feature" case.
    # We also test with `subquestion` not hidden, to ensure both ways
    # work correctly

    # Build the form first...
    topform = form_factory(slug="top")
    subformbranch = form_factory(slug="subform")
    subform = form_factory(slug="subformbranch")

    formquestion = question_factory(
        type=Question.TYPE_FORM,
        is_hidden=fq_is_hidden,
        sub_form=subform,
        slug="topformquestion",
    )
    formquestionbranch = question_factory(
        type=Question.TYPE_FORM,
        is_hidden="false",
        sub_form=subformbranch,
        slug="topformquestionbranch",
    )
    subquestion1 = question_factory(
        type=Question.TYPE_INTEGER, is_hidden="false", slug="subquestion1"
    )
    subquestion2 = question_factory(
        type=Question.TYPE_INTEGER, is_hidden="false", slug="subquestion2"
    )
    depquestion1 = question_factory(
        type=Question.TYPE_INTEGER,
        is_required="'subquestion1'|answer=='blah'",
        slug="depquestion1",
    )
    depquestion2 = question_factory(
        type=Question.TYPE_INTEGER,
        is_required="'subquestion2'|answer=='bluh'",
        slug="depquestion2",
    )
    form_question_factory(form=topform, question=formquestion)
    form_question_factory(form=topform, question=formquestionbranch)
    form_question_factory(form=topform, question=depquestion1)

    form_question_factory(form=subformbranch, question=depquestion2)

    form_question_factory(form=subform, question=subquestion1)
    form_question_factory(form=subform, question=subquestion2)

    # ... then build the document
    topdoc = document_factory(form=topform)
    answer_factory(document=topdoc, question=subquestion1, value="blah")
    answer_factory(document=topdoc, question=subquestion2, value="bluh")

    validator = validators.DocumentValidator()

    if fq_is_hidden == "true":
        # parent is hidden, required question cannot be
        # shown, thus is implicitly hidden
        validator.validate(topdoc, info)
        assert True  # above did not fail
    else:
        with pytest.raises(validators.CustomValidationError):
            validator.validate(topdoc, info)


def test_reference_missing_question(
    db, form_question_factory, form_factory, question_factory, document_factory, info
):
    topform = form_factory(slug="top")

    depquestion = question_factory(
        type=Question.TYPE_INTEGER, is_hidden="'subquestion-missing'|answer=='blah'"
    )
    form_question_factory(form=topform, question=depquestion)

    # ... then build the document
    topdoc = document_factory(form=topform)
    validator = validators.DocumentValidator()

    with pytest.raises(QuestionMissing):
        validator.validate(topdoc, info)


@pytest.mark.parametrize(
    "question,expr,expectation,features",
    [
        ("sub_question", "info.form == 'sub_form'", True, "subform"),
        ("sub_question", "info.formMeta.level == 1", True, "subform"),
        ("sub_question", "info.formMeta['is-top-form']", False, "subform"),
        ("sub_question", "info.formMeta['non-existent-key'] == null", True, "subform"),
        ("sub_question", "info.parent.form == 'top_form'", True, "subform"),
        ("sub_question", "info.parent.formMeta.level == 0", True, "subform"),
        ("sub_question", "info.parent.formMeta['is-top-form']", True, "subform"),
        ("column", "info.parent.form == 'top_form'", True, "table"),
        ("column", "info.parent.formMeta.level == 0", True, "table"),
        ("column", "info.parent.formMeta['is-top-form']", True, "table"),
        ("column", "info.root.form == 'top_form'", True, "table"),
        ("column", "info.root.formMeta.level == 0", True, "table"),
        ("column", "info.root.formMeta['is-top-form']", True, "table"),
    ],
)
def test_new_jexl_expressions(
    question, expr, expectation, features, info, form_and_document
):
    """Evaluate a JEXL expression in the context of a full document.

    The given JEXL expression is evaluated in the context of the given
    question within a structured document. The expression's value (boolean)
    is then returned. The document is generated with the following structure:

    * form: top_form
       * question: top_question
       * question: table [enabled by putting 'table' in features param]
           * row_form: row_form
               * question: column
       * question: form_question [enabled by putting 'subform' in features param]
           * sub_form: sub_form
               * question: sub_question
    """

    use_table = "table" in features
    use_subform = "subform" in features

    form, document, questions, answers = form_and_document(
        use_table=use_table, use_subform=use_subform
    )

    # expression test method: we delete an answer and set it's is_hidden
    # to an expression to be tested. If the expression evaluates to True,
    # we won't have a ValidationError.
    answers[question].delete()
    questions[question].is_hidden = expr
    questions[question].save()

    def do_check():
        validator = validators.DocumentValidator()
        try:
            validator.validate(document, info)
            return True
        except validators.CustomValidationError:  # pragma: no cover
            return False

    assert do_check() == expectation


def test_answer_transform_on_hidden_question(info, form_and_document):
    # Bug test: When a JEXL expression references two answers, and one of them
    # is hidden, the `answer` transform must return None for the hidden question
    # to ensure correct evaluation.

    form, document, questions, answers = form_and_document(
        use_table=True, use_subform=True
    )

    # the "sub_question" question is hidden (no magic). This means
    # it's "answer" transform should always return None
    questions["sub_question"].is_hidden = "true"
    questions["sub_question"].is_required = "true"
    questions["sub_question"].save()

    # sub_question and top_question are referenced in a JEXL expression later on
    answers["sub_question"].value = "hello"
    answers["sub_question"].save()
    answers["top_question"].value = "xyz"
    answers["top_question"].save()

    # the `column` question is used to evaluate the expression via `is_required`:
    # The required state depends on two questions, so we can check the expression's
    # result by checking whether the validator throws an exception.
    answers["column"].delete()

    # expression references two other questions, so it will still be evaluated
    # even if one question is hidden.
    # This specific expression evaluates to `True` only if the `answer` transform
    # works correctly
    questions[
        "column"
    ].is_required = "'sub_question'|answer == null && 'top_question'|answer=='xyz'"
    questions["column"].is_hidden = "false"
    questions["column"].save()

    validator = validators.DocumentValidator()
    with pytest.raises(validators.CustomValidationError):
        validator.validate(document, info)

    # Counter-Test to the `expect_fail` case: Ensure the answer transform still works
    questions["column"].is_required = "'sub_question'|answer == 'hello'"
    questions["column"].is_hidden = "false"
    questions["column"].save()

    assert validator.validate(document, info) is None


@pytest.mark.parametrize(
    "question_type,expected_value",
    [
        (Question.TYPE_MULTIPLE_CHOICE, []),
        (Question.TYPE_INTEGER, None),
        (Question.TYPE_FLOAT, None),
        (Question.TYPE_DATE, None),
        (Question.TYPE_CHOICE, None),
        (Question.TYPE_TEXTAREA, None),
        (Question.TYPE_TEXT, None),
        (Question.TYPE_TABLE, []),
        (Question.TYPE_FILE, None),
        (Question.TYPE_DYNAMIC_CHOICE, None),
        (Question.TYPE_DYNAMIC_MULTIPLE_CHOICE, []),
        # Those should not appear in a JEXL answer transform
        # (Question.TYPE_FORM,None),
        # (Question.TYPE_STATIC,None),
    ],
)
def test_answer_transform_on_hidden_question_types(
    info,
    form_and_document,
    document_factory,
    answer_factory,
    question_type,
    expected_value,
):
    form, document, questions, answers = form_and_document(
        use_table=True, use_subform=True
    )
    table = questions["table"]
    row_form = table.row_form
    row_doc = document_factory(form=row_form)
    answer_factory(document=row_doc, question=questions["column"])
    answers["table"].documents.add(row_doc)

    questions["form"].is_hidden = (
        f"'top_question'|answer == {expected_value}"
        " && 'table'|answer|mapby('column')[0]"
        " && 'table'|answer|mapby('column')[1]"
    )
    questions["form"].save()

    questions["top_question"].is_hidden = "true"
    questions["top_question"].type = question_type
    questions["top_question"].save()

    qj = QuestionJexl(
        {
            "document": document,
            "answers": answers,
            "form": form,
            "structure": structure.FieldSet(document, document.form),
        }
    )

    field = structure.Field(document, form, questions["form"])
    assert qj.is_hidden(field)


@pytest.mark.parametrize(
    "jexl_field,expr",
    [
        ("is_required", "('column'|answer == 5)"),
        ("is_hidden", "('column'|answer > 10)"),
    ],
)
def test_answer_transform_in_tables(
    info,
    form_and_document,
    form_question_factory,
    jexl_field,
    expr,
    answer_document_factory,
):

    form, document, questions, answers = form_and_document(
        use_table=True, use_subform=False
    )
    column_a = answers["column"]
    column_a.value = 11
    column_a.save()

    table_question = questions["table"]

    col2_question = form_question_factory(
        **{
            "form": table_question.row_form,
            "question__slug": "column2",
            "question__is_hidden": "false",
            "question__is_required": "true",
            # this overrwrites above "default values"
            f"question__{jexl_field}": expr,
        }
    ).question
    assert getattr(col2_question, jexl_field) == expr

    table_answer = answers["table"]

    row2_doc = answer_document_factory(
        answer=table_answer, document__form=table_question.row_form, sort=10
    ).document

    # Second row has an answer that triggers the transform and implies a validation error.
    # This will wrongfully succeed if the answer value comes from the wrong row
    row2_doc.answers.create(question_id="column", value=5)

    validator = validators.DocumentValidator()
    # we expect this to fail, as in the second row, the 'column' answer is 5,
    # so 'column2' should be required
    with pytest.raises(validators.CustomValidationError):
        validator.validate(document, info)


def test_answer_transform_hidden_table_cell(
    info,
    form_and_document,
    form_question_factory,
):
    form, document, questions, answers = form_and_document(
        use_table=True, use_subform=False
    )
    column_a = answers["column"]
    column_a.value = 11.0
    column_a.save()

    table_question = questions["table"]

    # create column that is always hidden
    col2_question = form_question_factory(
        form=table_question.row_form,
        question__slug="column2",
        question__is_hidden="true",
        question__is_required="false",
    ).question

    # create question that depends on the value of the previously created hidden
    # column
    form_question_factory(
        form=form,
        question__is_hidden=f"!('yes' in '{table_question.slug}'|answer|mapby('{col2_question.slug}'))",
        question__is_required="true",
    )

    # answer the hidden column with the value that would make the dependant
    # question visible
    row = answers["table"].documents.first()
    row.answers.create(question_id=col2_question.slug, value="yes")

    # validation should suceed since the check question should be hidden even
    # though the dependant column has the proper value but is hidden
    validator = validators.DocumentValidator()
    validator.validate(document, info)

    assert True


def test_is_hidden_neighboring_table(
    info, form_question_factory, form_factory, form_and_document
):
    """
    Test a more complicated, nested setup.

    The structure looks like this:

    * form: top_form
       * question: top_question
       * question: form_question
           * sub_form: sub_form
               * question: sub_question
           * question: table
               * row_form: row_form
                   * question: column

       * question: neighbor_form_question
           * neighbor_form
               * question: neighbor_sub_question

    The is_hidden on neighbor_sub_question references the column question.
    """

    form, document, questions, answers = form_and_document(
        use_table=True, use_subform=True
    )

    # delete table answer documents
    answers["table"].documents.first().delete()
    assert not models.Document.objects.filter(form_id="row_form").exists()
    assert not models.Answer.objects.filter(question_id="column").exists()

    form_q = questions["form"]
    table_q = questions["table"]

    table_q.forms.remove(form)
    table_q.forms.add(form_q.sub_form)

    neighbor_form = form_factory()
    neighbor_sub_question = form_question_factory(
        form=neighbor_form,
        question__is_hidden="!('foo' in 'table'|answer|mapby('column'))",
        question__is_required=True,
    ).question

    form_question_factory(
        form=form, question__type=Question.TYPE_FORM, question__sub_form=neighbor_form
    )

    qj = QuestionJexl(
        {
            "document": document,
            "answers": answers,
            "form": form,
            "structure": structure.FieldSet(document, document.form),
        }
    )

    field = structure.Field(document, form, neighbor_sub_question)

    assert qj.is_hidden(field)

    validator = validators.DocumentValidator()
    assert neighbor_sub_question not in validator.visible_questions(document)

    with pytest.raises(validators.CustomValidationError):
        validator.validate(document, info)



import logging
import sys

import pytest
import requests
from django.conf import settings
from django.core import management
from django.db import connection
from minio import Minio
from psycopg2 import OperationalError
from requests.models import Response
from watchman.decorators import check as watchman_check_decorator

from caluma.caluma_form import storage_clients


@pytest.fixture
def disable_logs():
    """Disable logs for health checks."""
    urllib3_logger = logging.getLogger("urllib3")
    watchman_logger = logging.getLogger("watchman")

    urllib3_level = urllib3_logger.level
    watchman_disabled = watchman_logger.disabled
    urllib3_logger.setLevel(logging.ERROR)
    watchman_logger.disabled = True

    yield

    urllib3_logger.setLevel(urllib3_level)
    watchman_logger.disabled = watchman_disabled


@pytest.fixture
def track_errors(mocker):
    """Write stacktrace to stderr to enable error checking in tests.

    Patches django-watchman check decorator.
    """

    def track_error_check_decorator(func):
        def wrapped(*args, **kwargs):
            response = watchman_check_decorator(func)(*args, **kwargs)

            if response.get("ok") is False:
                sys.stderr.write(response["stacktrace"])
            elif not response.get("ok"):
                for value in response.values():
                    if value.get("ok") is False:
                        sys.stderr.write(value["stacktrace"])

            return response

        return wrapped

    mocker.patch("watchman.decorators.check", new=track_error_check_decorator)

    yield


@pytest.fixture
def db_broken_connection(transactional_db):
    """DB fixture with broken connection.

    Cap DB connection by trying to re-connect with faulty connection details.
    Changes host settings and restores them after use.
    """
    # overwrite DB host setting and re-connect
    try:
        connection.close()
        prev_settings = settings.DATABASES["default"]["HOST"]
        settings.DATABASES["default"]["HOST"] = "not-db"
        connection.connect()
    except OperationalError:
        # ignore connection error
        pass

    yield

    # restore original DB settings
    connection.close()
    settings.DATABASES["default"]["HOST"] = prev_settings
    connection.connect()


def switch_user(test=False):
    """Re-connect to database with either user 'tester' or 'caluma'."""
    # set username and password
    user = "tester" if test else "caluma"
    password = "test" if test else "caluma"

    # connect new user to database
    connection.close()
    settings.DATABASES["default"]["USER"] = user
    settings.DATABASES["default"]["PASSWORD"] = password
    connection.connect()


@pytest.fixture
def db_user_tester(transactional_db):
    """Create new user 'tester' with privileges necessary for health checks.

    User is removed after use.
    """
    # get db connection
    cursor = connection.cursor()

    # create test user (with all necessary privileges)
    cursor.execute("CREATE USER tester WITH PASSWORD 'test';")
    cursor.execute("GRANT ALL ON caluma_logging_accesslog TO tester;")
    cursor.execute("GRANT ALL ON caluma_form_file TO tester;")
    cursor.execute(
        "GRANT USAGE, SELECT ON SEQUENCE caluma_logging_accesslog_id_seq TO tester;"
    )
    cursor.execute("GRANT ALL ON django_migrations TO tester;")

    yield

    # get db connection
    cursor = connection.cursor()

    # revoke privileges and remove user
    cursor.execute("REVOKE ALL ON caluma_logging_accesslog FROM tester;")
    cursor.execute("REVOKE ALL ON caluma_form_file FROM tester;")
    cursor.execute(
        "REVOKE USAGE, SELECT ON SEQUENCE caluma_logging_accesslog_id_seq FROM tester;"
    )
    cursor.execute("REVOKE ALL ON django_migrations FROM tester;")
    cursor.execute("DROP USER tester;")


@pytest.fixture
def db_privileges_working(db_user_tester):
    """DB fixture with all necessary privileges for health checks.

    Create new test user with necessary privileges and connect to database.
    Remove test user and reconnect with original database user after use.
    """
    # switch to new test user
    switch_user(test=True)

    yield

    # switch back to caluma user
    switch_user()


@pytest.fixture
def db_read_error(db_user_tester):
    """DB fixture with revoked read privileges on caluma_logging_accesslog.

    Create new test user without necessary read privileges for table used by
    health check and connect to database.
    Remove test user and reconnect with original database user after use.
    """
    # revoke select rights on accesslog table
    cursor = connection.cursor()
    cursor.execute("REVOKE SELECT ON caluma_logging_accesslog FROM tester;")

    # switch to new test user
    switch_user(test=True)

    yield

    # switch back to caluma user
    switch_user()

    # grant select rights on accesslog table such that user can be removed
    cursor = connection.cursor()
    cursor.execute("GRANT SELECT ON caluma_logging_accesslog TO tester;")


@pytest.fixture
def db_write_error(db_user_tester):
    """DB fixture with revoked write privileges on caluma_logging_accesslog.

    Create new test user without necessary write privileges for table used by
    health check and connect to database.
    Remove test user and reconnect with original database user after use.
    """
    # revoke insert, update and delete privileges on accesslog table
    cursor = connection.cursor()
    cursor.execute(
        "REVOKE INSERT,UPDATE,DELETE ON caluma_logging_accesslog FROM tester;"
    )

    # switch to new test user
    switch_user(test=True)

    yield

    # switch back to caluma user
    switch_user()

    # grant insert, update, delete rights on accesslog table
    # such that user can be removed
    cursor = connection.cursor()
    cursor.execute("GRANT INSERT,UPDATE,DELETE ON caluma_logging_accesslog TO tester;")


@pytest.fixture
def minio_mock_working(minio_mock, mocker):
    """Provide working minio mock for health checks.

    Based on existing minio_mock.
    Patches requests.get and requests.put to provide necessary responses for
    health check.
    """
    # determine fixed UUID to be used during tests
    uuid = "7c5ad424-e351-4db8-ba37-bbace45d0e0f"
    mocker.patch("uuid.uuid4", return_value=uuid)

    # mock upload response
    upload_response = Response()
    upload_response.status_code = 200
    mocker.patch("requests.put", return_value=upload_response)

    # mock download response
    download_response = Response()
    download_response.status_code = 200
    download_response._content = (
        b"--ffc537c8d0da305912d6440576ac0d75\r\n"
        b'Content-Disposition: form-data; name="file"; '
        b'filename="healthz_tmp-file.txt"\r\n\r\n'
        b"Test\nTest\n\r\n"
        b"--ffc537c8d0da305912d6440576ac0d75--\r\n"
    )
    mocker.patch("requests.get", return_value=download_response)

    # change object name of existing stat response
    stat_response = Minio.stat_object.return_value
    stat_response._object_name = "healthz_tmp-object_" + uuid

    # update responses compatible with media storage service health check
    Minio.stat_object.side_effect = [stat_response, None]


@pytest.fixture
def minio_broken_connection():
    """Provide new minio client with faulty connection details.

    Change minio settings to generate new client unable to connect to minio
    instance.
    Restore original settings and client after use.
    """
    # store previous client
    prev_client = storage_clients.client

    # reconnect with new client (with invalid host settings)
    prev_settings = settings.MINIO_STORAGE_ENDPOINT
    settings.MINIO_STORAGE_ENDPOINT = "not-minio:9000"
    storage_clients.client = storage_clients.Minio()

    yield

    # restore previous client
    settings.MINIO_STORAGE_ENDPOINT = prev_settings
    storage_clients.client = prev_client


@pytest.fixture
def minio_not_configured():
    """Fixture with no connection to minio instance.

    Remove minio client to ensure no connection is available.
    """
    # store previous client
    prev_client = storage_clients.client

    # reset client
    storage_clients.client = None

    yield

    # restore previous client
    storage_clients.client = prev_client


@pytest.fixture
def minio_mock_read_error(minio_mock, mocker):
    """Produce exception when downloading file or stat-ing object.

    This scenario might occur when there is a internal minio issue or if a
    previous upload isn't successful (which doesn't issue exception).
    """
    # raise exception instead of handling request
    def download_side_effect(*args, **kwargs):  # pragma: no cover
        raise requests.exceptions.RequestException("GET request failed")

    def stat_side_effect(*args, **kwargs):
        raise Exception("stat_object failed")

    # determine fixed UUID to be used during tests
    uuid = "7c5ad424-e351-4db8-ba37-bbace45d0e0f"
    mocker.patch("uuid.uuid4", return_value=uuid)

    # mock upload response
    upload_response = Response()
    upload_response.status_code = 200
    mocker.patch("requests.put", return_value=upload_response)

    # mock erroneous download response
    download_response = Response()
    download_response.status_code = 500
    mocker.patch(
        "requests.get", return_value=download_response, side_effect=download_side_effect
    )

    # raise exception
    Minio.stat_object.side_effect = stat_side_effect


@pytest.fixture
def minio_mock_write_error(minio_mock, mocker):
    """Produce exception when uploading file.

    This scenario might occur when there is a internal minio issue.
    """
    # raise exception instead of handling request
    def upload_side_effect(*args, **kwargs):
        raise requests.exceptions.RequestException("PUT request failed")

    # determine fixed UUID to be used during tests
    uuid = "7c5ad424-e351-4db8-ba37-bbace45d0e0f"
    mocker.patch("uuid.uuid4", return_value=uuid)

    # mock upload response
    upload_response = Response()
    upload_response.status_code = 500
    mocker.patch(
        "requests.put", return_value=upload_response, side_effect=upload_side_effect
    )

    # mock download response
    download_response = Response()
    download_response.status_code = 200
    download_response._content = (
        b"--ffc537c8d0da305912d6440576ac0d75\r\n"
        b'Content-Disposition: form-data; name="file"; '
        b'filename="healthz_tmp-file.txt"\r\n\r\n'
        b"Test\nTest\n\r\n"
        b"--ffc537c8d0da305912d6440576ac0d75--\r\n"
    )
    mocker.patch("requests.get", return_value=download_response)

    # change object name of existing stat response
    stat_response = Minio.stat_object.return_value
    stat_response._object_name = "healthz_tmp-object_" + uuid

    # update responses compatible with media storage service health check
    Minio.stat_object.side_effect = [stat_response, None]


@pytest.fixture
def roll_back_migrations():
    """Roll back database migrations to ensure unapplied migrations exist."""
    # undo applied migrations for app contenttypes
    management.call_command("migrate", "contenttypes", "zero")

    yield

    # re-apply migrations for app contenttypes
    management.call_command("migrate", "contenttypes")



def test_base_model_created_by_user_filter(db, work_item_factory, schema_executor):
    work_item_factory(created_by_user="Winston Smith")
    work_item_factory(created_by_user="Martin")

    query = """
            query WorkItems($createdByUser: String!) {
              allWorkItems(createdByUser: $createdByUser) {
                edges {
                  node {
                    createdByUser
                    createdByGroup
                  }
                }
              }
            }
        """

    result = schema_executor(query, variable_values={"createdByUser": "Winston Smith"})

    assert not result.errors
    assert len(result.data["allWorkItems"]["edges"]) == 1
    assert (
        result.data["allWorkItems"]["edges"][0]["node"]["createdByUser"]
        == "Winston Smith"
    )


def test_base_model_created_by_group_filter(db, work_item_factory, schema_executor):
    work_item_factory(created_by_group="Oceania")
    work_item_factory(created_by_group="Eurasia")

    query = """
                query WorkItems($createdByGroup: String!) {
                  allWorkItems(createdByGroup: $createdByGroup) {
                    edges {
                      node {
                        createdByUser
                        createdByGroup
                      }
                    }
                  }
                }
            """

    result = schema_executor(query, variable_values={"createdByGroup": "Oceania"})

    assert not result.errors
    assert len(result.data["allWorkItems"]["edges"]) == 1
    assert (
        result.data["allWorkItems"]["edges"][0]["node"]["createdByGroup"] == "Oceania"
    )



from ..caluma_core.types import Node
from ..caluma_core.visibilities import BaseVisibility, filter_queryset_for


class Authenticated(BaseVisibility):
    """Only allow authenticated users to read nodes."""

    @filter_queryset_for(Node)
    def filter_queryset_for_all(self, node, queryset, info):
        if info.context.user.is_authenticated:
            return queryset
        return queryset.none()


class CreatedByGroup(BaseVisibility):
    """User may only read nodes of created by group it belongs to."""

    @filter_queryset_for(Node)
    def filter_queryset_for_all(self, node, queryset, info):
        groups = info.context.user.groups
        return queryset.filter(created_by_group__in=groups)



# Generated by Django 2.2.15 on 2020-09-01 14:15

import caluma.caluma_core.models
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("caluma_workflow", "0024_auto_20200901_1026")]

    operations = [
        migrations.AlterField(
            model_name="historicalworkitem",
            name="status",
            field=caluma.caluma_core.models.ChoicesCharField(
                choices=[
                    ("ready", "Work item is ready to be processed."),
                    ("completed", "Work item is done."),
                    ("canceled", "Work item is canceled."),
                    ("skipped", "Work item is skipped."),
                    ("suspended", "Work item is suspended."),
                ],
                db_index=True,
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="workitem",
            name="status",
            field=caluma.caluma_core.models.ChoicesCharField(
                choices=[
                    ("ready", "Work item is ready to be processed."),
                    ("completed", "Work item is done."),
                    ("canceled", "Work item is canceled."),
                    ("skipped", "Work item is skipped."),
                    ("suspended", "Work item is suspended."),
                ],
                db_index=True,
                max_length=50,
            ),
        ),
    ]



# Generated by Django 2.2.18 on 2021-03-04 14:11

from django.db import migrations, models
from django.db.models.expressions import F


def initial_modified_properties(apps, schema_editor):
    for model in [
        "Case",
        "Flow",
        "Task",
        "TaskFlow",
        "WorkFlow",
        "WorkItem",
    ]:
        Model = apps.get_model(f"caluma_workflow.{model}")
        Model.objects.update(
            modified_by_user=F("created_by_user"),
            modified_by_group=F("created_by_group"),
        )


class Migration(migrations.Migration):

    dependencies = [
        ("caluma_workflow", "0026_auto_20200901_1435"),
    ]

    operations = [
        migrations.AddField(
            model_name="case",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="case",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="flow",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="flow",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicalcase",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicalcase",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicalflow",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicalflow",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicaltask",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicaltask",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicaltaskflow",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicaltaskflow",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicalworkflow",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicalworkflow",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicalworkitem",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicalworkitem",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="task",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="task",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="taskflow",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="taskflow",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="workflow",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="workflow",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="workitem",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="workitem",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.RunPython(initial_modified_properties, migrations.RunPython.noop),
    ]



import pytest
from django.utils import translation
from graphql_relay import to_global_id

from ...caluma_core.tests import extract_serializer_input_fields
from .. import models
from ..api import copy_form
from ..serializers import SaveFormSerializer


@pytest.mark.parametrize(
    "form__description,form__name,question__type",
    [("First result", "1st", models.Question.TYPE_FLOAT)],
)
def test_query_all_forms(
    db,
    snapshot,
    form,
    form_factory,
    form_question,
    form_question_factory,
    question,
    schema_executor,
):
    form_factory(name="3rd", description="Second result")
    form_factory(name="2nd", description="Second result")
    form_question_factory(form=form)

    query = """
        query AllFormsQuery($name: String, $question: String, $orderBy: [FormOrdering]) {
          allForms(name: $name, orderBy: $orderBy) {
            edges {
              node {
                id
                slug
                name
                description
                meta
                questions(search: $question) {
                  edges {
                    node {
                      id
                      slug
                      label
                    }
                  }
                }
              }
            }
          }
        }
    """

    result = schema_executor(
        query,
        variable_values={
            "question": question.label,
            "orderBy": ["NAME_ASC", "CREATED_AT_ASC"],
        },
    )

    assert not result.errors
    snapshot.assert_match(result.data)


@pytest.mark.parametrize("language_code", ("en", "de"))
@pytest.mark.parametrize("form__description", ("some description text", ""))
def test_save_form(db, snapshot, form, settings, schema_executor, language_code):
    query = """
        mutation SaveForm($input: SaveFormInput!) {
          saveForm(input: $input) {
            form {
              id
              slug
              name
              meta
            }
            clientMutationId
          }
        }
    """

    inp = {"input": extract_serializer_input_fields(SaveFormSerializer, form)}
    with translation.override(language_code):
        result = schema_executor(query, variable_values=inp)

    assert not result.errors
    snapshot.assert_match(result.data)


def test_save_form_created_as_admin_user(db, form, admin_schema_executor, admin_user):
    query = """
        mutation SaveForm($input: SaveFormInput!) {
          saveForm(input: $input) {
            form {
              createdByUser
            }
          }
        }
    """

    inp = {"input": extract_serializer_input_fields(SaveFormSerializer, form)}
    form.delete()  # test creation of form
    result = admin_schema_executor(query, variable_values=inp)

    assert not result.errors
    assert result.data["saveForm"]["form"]["createdByUser"] == admin_user.username


@pytest.mark.parametrize("form__meta", [{"meta": "set"}])
def test_copy_form(db, form, form_question_factory, schema_executor):
    form_question_factory.create_batch(5, form=form)
    query = """
        mutation CopyForm($input: CopyFormInput!) {
          copyForm(input: $input) {
            form {
              slug
            }
            clientMutationId
          }
        }
    """

    inp = {"input": {"source": form.pk, "slug": "new-form", "name": "Test Form"}}
    result = schema_executor(query, variable_values=inp)

    assert not result.errors

    form_slug = result.data["copyForm"]["form"]["slug"]
    assert form_slug == "new-form"
    new_form = models.Form.objects.get(pk=form_slug)
    assert new_form.name == "Test Form"
    assert new_form.meta == form.meta
    assert new_form.source == form
    assert list(
        models.FormQuestion.objects.filter(form=new_form).values("question")
    ) == list(models.FormQuestion.objects.filter(form=form).values("question"))


@pytest.mark.parametrize("form__meta", [{"meta": "set"}])
def test_copy_form_api(db, form, form_question_factory, schema_executor):
    form_question_factory.create_batch(5, form=form)

    new_form = copy_form(source=form, slug="new-form", name="Test Form")

    assert new_form.pk == "new-form"
    assert new_form.name == "Test Form"
    assert new_form.meta == form.meta
    assert new_form.source == form
    assert list(
        models.FormQuestion.objects.filter(form=new_form).values("question")
    ) == list(models.FormQuestion.objects.filter(form=form).values("question"))


def test_add_form_question(db, form, question, form_question_factory, schema_executor):
    form_questions = form_question_factory.create_batch(5, form=form)

    # initialize sorting keys
    for idx, form_question in enumerate(form_questions):
        form_question.sort = idx + 1
        form_question.save()

    query = """
        mutation AddFormQuestion($input: AddFormQuestionInput!) {
          addFormQuestion(input: $input) {
            form {
              questions {
                edges {
                  node {
                    slug
                  }
                }
              }
            }
            clientMutationId
          }
        }
    """

    result = schema_executor(
        query,
        variable_values={
            "input": {
                "form": to_global_id(type(form).__name__, form.pk),
                "question": to_global_id(type(question).__name__, question.pk),
            }
        },
    )

    assert not result.errors
    questions = result.data["addFormQuestion"]["form"]["questions"]["edges"]
    assert len(questions) == 6
    assert questions[-1]["node"]["slug"] == question.slug


def test_remove_form_question(
    db, form, form_question, question, snapshot, schema_executor
):
    query = """
        mutation RemoveFormQuestion($input: RemoveFormQuestionInput!) {
          removeFormQuestion(input: $input) {
            form {
              questions {
                edges {
                  node {
                    slug
                  }
                }
              }
            }
            clientMutationId
          }
        }
    """

    result = schema_executor(
        query,
        variable_values={
            "input": {
                "form": to_global_id(type(form).__name__, form.pk),
                "question": to_global_id(type(question).__name__, question.pk),
            }
        },
    )

    assert not result.errors
    snapshot.assert_match(result.data)


def test_reorder_form_questions(db, form, form_question_factory, schema_executor):
    form_question_factory.create_batch(2, form=form)

    query = """
        mutation ReorderFormQuestions($input: ReorderFormQuestionsInput!) {
          reorderFormQuestions(input: $input) {
            form {
              questions {
                edges {
                  node {
                    slug
                  }
                }
              }
            }
            clientMutationId
          }
        }
    """

    question_ids = (
        form.questions.order_by("slug").reverse().values_list("slug", flat=True)
    )
    result = schema_executor(
        query,
        variable_values={
            "input": {
                "form": to_global_id(type(form).__name__, form.pk),
                "questions": [
                    to_global_id(models.Question.__name__, question_id)
                    for question_id in question_ids
                ],
            }
        },
    )

    assert not result.errors
    result_questions = [
        question["node"]["slug"]
        for question in result.data["reorderFormQuestions"]["form"]["questions"][
            "edges"
        ]
    ]

    assert result_questions == list(question_ids)


def test_reorder_form_questions_invalid_question(
    db, form, question_factory, schema_executor
):

    invalid_question = question_factory()

    query = """
        mutation ReorderFormQuestions($input: ReorderFormQuestionsInput!) {
          reorderFormQuestions(input: $input) {
            form {
              questions {
                edges {
                  node {
                    slug
                  }
                }
              }
            }
            clientMutationId
          }
        }
    """

    result = schema_executor(
        query,
        variable_values={
            "input": {
                "form": to_global_id(type(form).__name__, form.pk),
                "questions": [
                    to_global_id(type(models.Question).__name__, invalid_question.slug)
                ],
            }
        },
    )

    assert result.errors


def test_reorder_form_questions_duplicated_question(
    db, form, question, form_question, schema_executor
):

    query = """
        mutation ReorderFormQuestions($input: ReorderFormQuestionsInput!) {
          reorderFormQuestions(input: $input) {
            form {
              questions {
                edges {
                  node {
                    slug
                  }
                }
              }
            }
            clientMutationId
          }
        }
    """

    result = schema_executor(
        query,
        variable_values={
            "input": {
                "form": to_global_id(type(form).__name__, form.pk),
                "questions": [question.slug, question.slug],
            }
        },
    )
    assert result.errors



# Generated by Django 2.2.16 on 2020-11-09 06:23

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0035_historicalanswer_history_question_type")]

    operations = [
        migrations.AddField(
            model_name="historicalquestion",
            name="default_answer",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="caluma_form.Answer",
            ),
        ),
        migrations.AddField(
            model_name="question",
            name="default_answer",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="caluma_form.Answer",
            ),
        ),
        migrations.AlterField(
            model_name="answer",
            name="document",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="answers",
                to="caluma_form.Document",
            ),
        ),
    ]



import pytest
from django.core.exceptions import ImproperlyConfigured
from django.db.models import IntegerField
from django.http import Http404
from rest_framework import exceptions
from rest_framework.serializers import CharField, Serializer

from .. import models, mutation, permissions, serializers, types, validations
from .fake_model import get_fake_model


def test_missing_mutation_serializer_class():
    with pytest.raises(Exception) as exc:

        class MyMutation(mutation.Mutation):
            pass

    assert str(exc.value) == "serializer_class is required for the Mutation"


def test_invalid_mutation_model_operations(db):
    class MySerializer(serializers.ModelSerializer):
        class Meta:
            model = get_fake_model()
            fields = "__all__"

    with pytest.raises(Exception) as exc:

        class MyMutation(mutation.Mutation):
            class Meta:
                serializer_class = MySerializer
                model_operations = ["Add"]

    assert "model_operations" in str(exc.value)


def test_invalid_mutation_update_mutate_and_get_payload(db, info):
    FakeModel = get_fake_model(model_base=models.UUIDModel)

    class FakeModelObjectType(types.DjangoObjectType):
        class Meta:
            model = FakeModel

        @classmethod
        def get_queryset(cls, queryset, info):
            # enforce that nothing is visible
            return queryset.none()

    class MySerializer(serializers.ModelSerializer):
        class Meta:
            model = get_fake_model()
            fields = "__all__"

    class InvalidModelMutation(mutation.Mutation):
        class Meta:
            serializer_class = MySerializer
            return_field_type = FakeModelObjectType
            model_operations = ["update"]

    with pytest.raises(Exception) as exc:
        InvalidModelMutation.mutate_and_get_payload(None, info)

    assert '"id" required' in str(exc.value)


def test_mutation_mutate_and_get_payload_without_model(info):
    class MySerializer(Serializer):
        name = CharField()

        def create(self, validated_data):
            return validated_data

    class NoModelMutation(mutation.Mutation):
        class Meta:
            return_field_name = "test"
            serializer_class = MySerializer

    result = NoModelMutation.mutate_and_get_payload(None, info, name="test")
    assert type(result) == NoModelMutation


def test_mutation_mutate_and_get_payload_without_permission(db, info):
    class MySerializer(serializers.ModelSerializer):
        class Meta:
            model = get_fake_model()
            fields = "__all__"

    class NoPermission(permissions.BasePermission):
        def has_permission(self, mutation, info):
            return False

    class MyMutation(mutation.Mutation):
        permission_classes = [NoPermission]

        class Meta:
            serializer_class = MySerializer

    with pytest.raises(exceptions.PermissionDenied):
        MyMutation.mutate_and_get_payload(None, info)


def test_mutation_mutate_and_get_payload_without_object_permission(db, info):
    FakeModel = get_fake_model()
    instance = FakeModel.objects.create()

    class FakeModelObjectType(types.DjangoObjectType):
        class Meta:
            model = FakeModel

    class MySerializer(serializers.ModelSerializer):
        class Meta:
            model = FakeModel
            fields = "__all__"

    class NoObjectPermission(permissions.BasePermission):
        def has_object_permission(self, mutation, info, instance):
            return False

    class MyMutation(mutation.Mutation):
        permission_classes = [NoObjectPermission]

        class Meta:
            serializer_class = MySerializer
            return_field_type = FakeModelObjectType

    with pytest.raises(exceptions.PermissionDenied):
        MyMutation.mutate_and_get_payload(None, info, id=str(instance.pk))


def test_mutation_mutate_and_get_payload_permission_classes_improperly_configured(
    db, info
):
    class MySerializer(serializers.ModelSerializer):
        class Meta:
            model = get_fake_model()
            fields = "__all__"

    class MyMutation(mutation.Mutation):
        permission_classes = None

        class Meta:
            serializer_class = MySerializer

    with pytest.raises(ImproperlyConfigured):
        MyMutation.mutate_and_get_payload(None, info)


def test_mutation_mutate_and_get_payload_validation_classes_improperly_configured(
    db, info
):
    FakeModel = get_fake_model()

    class MySerializer(serializers.ModelSerializer):
        validation_classes = None

        class Meta:
            model = FakeModel
            fields = "__all__"

    class CustomNode(types.DjangoObjectType):
        class Meta:
            model = FakeModel

    class MyMutation(mutation.Mutation):
        class Meta:
            serializer_class = MySerializer

    with pytest.raises(ImproperlyConfigured):
        MyMutation.mutate_and_get_payload(None, info)


def test_mutation_mutate_and_get_payload_validation_classes_custom_validation(
    db, info, history_mock
):
    FakeModel = get_fake_model(
        model_base=models.UUIDModel, fields={"testnum": IntegerField(null=True)}
    )

    class CustomValidation(validations.BaseValidation):
        def validate(self, mutation, data, info):
            data["testnum"] = 1
            return data

    class MySerializer(serializers.ModelSerializer):
        validation_classes = [CustomValidation]

        class Meta:
            model = FakeModel
            fields = "__all__"

    class CustomNode(types.DjangoObjectType):
        class Meta:
            model = FakeModel

    class MyMutation(mutation.Mutation):
        class Meta:
            serializer_class = MySerializer

    MyMutation.mutate_and_get_payload(None, info)
    assert FakeModel.objects.first().testnum == 1


def test_user_defined_primary_key_get_serializer_kwargs_not_allowed(
    db, info, history_mock
):
    """Test that user may not overwrite existing instance which is not visible."""
    FakeModel = get_fake_model(model_base=models.SlugModel)

    class FakeModelObjectType(types.DjangoObjectType):
        class Meta:
            model = FakeModel

        @classmethod
        def get_queryset(cls, queryset, info):
            # enforce that nothing is visible
            return queryset.none()

    class MySerializer(serializers.ModelSerializer):
        class Meta:
            model = FakeModel
            fields = "__all__"

    class MyMutation(mutation.UserDefinedPrimaryKeyMixin, mutation.Mutation):
        class Meta:
            serializer_class = MySerializer
            return_field_type = FakeModelObjectType

    FakeModel.objects.create(slug="test")

    with pytest.raises(Http404):
        MyMutation.get_serializer_kwargs(None, info, slug="test")


def test_user_defined_primary_key_get_serializer_kwargs_update_not_allowed(
    db, info, history_mock
):
    FakeModel = get_fake_model(model_base=models.SlugModel)

    class FakeModelObjectType(types.DjangoObjectType):
        class Meta:
            model = FakeModel

    class MySerializer(serializers.ModelSerializer):
        class Meta:
            model = FakeModel
            fields = "__all__"

    class MyMutation(mutation.UserDefinedPrimaryKeyMixin, mutation.Mutation):
        class Meta:
            serializer_class = MySerializer
            return_field_type = FakeModelObjectType
            model_operations = ["create"]

    FakeModel.objects.create(slug="test")

    with pytest.raises(exceptions.ValidationError):
        MyMutation.get_serializer_kwargs(None, info, slug="test")


def test_user_defined_primary_key_get_serializer_kwargs_create_not_allowed(db, info):
    FakeModel = get_fake_model(model_base=models.SlugModel)

    class FakeModelObjectType(types.DjangoObjectType):
        class Meta:
            model = FakeModel

    class MySerializer(serializers.ModelSerializer):
        class Meta:
            model = FakeModel
            fields = "__all__"

    class MyMutation(mutation.UserDefinedPrimaryKeyMixin, mutation.Mutation):
        class Meta:
            serializer_class = MySerializer
            return_field_type = FakeModelObjectType
            model_operations = ["update"]

    with pytest.raises(exceptions.ValidationError):
        MyMutation.get_serializer_kwargs(None, info, slug="test")



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-01-21 14:40
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_workflow", "0005_auto_20181228_1243")]

    operations = [
        migrations.AlterField(
            model_name="task",
            name="address_groups",
            field=models.TextField(
                blank=True,
                help_text="Group jexl returning what group(s) derived work items will be addressed to.",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="workflow",
            name="start",
            field=models.ForeignKey(
                help_text="First task of the workflow.",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="caluma_workflow.Task",
            ),
        ),
    ]
